# -*- coding: utf-8 -*-
"""Reconstruct magnetic distributions from given phasemaps.

This module reconstructs 3-dimensional magnetic distributions (as :class:`~pyramid.magdata.MagData`
objects) from a given set of phase maps (represented by :class:`~pyramid.phasemap.PhaseMap`
objects) by using several model based reconstruction algorithms which use the forward model
provided by :mod:`~pyramid.projector` and :mod:`~pyramid.phasemapper` and a priori knowledge of
the distribution.
So far, only a simple least square algorithm for known pixel locations for 2-dimensional problems
is implemented (:func:`~.reconstruct_simple_leastsq`), but more complex solutions are planned.

"""


import numpy as np

from scipy.optimize import leastsq

import pyramid.projector as pj
import pyramid.phasemapper as pm
from pyramid.magdata import MagData


def reconstruct_simple_leastsq(phase_map, mask, b_0=1):
    '''Reconstruct a magnetic distribution for a 2-D problem with known pixel locations.

    Parameters
    ----------
        phase_map : :class:`~pyramid.phasemap.PhaseMap`
            A :class:`~pyramid.phasemap.PhaseMap` object, representing the phase from which to
            reconstruct the magnetic distribution.
        mask : :class:`~numpy.ndarray` (N=3)
            A boolean matrix (or a matrix consisting of ones and zeros), representing the
            positions of the magnetized voxels in 3 dimensions.
        b_0 : float, optional
            The magnetic induction corresponding to a magnetization `M`\ :sub:`0` in T.
            The default is 1.
    Returns
    -------
        mag_data : :class:`~pyramid.magdata.MagData`
            The reconstructed magnetic distribution as a
            :class:`~pyramid.magdata.MagData` object.

    Notes
    -----
    Only works for a single phase_map, if the positions of the magnetized voxels are known and
    for slice thickness of 1 (constraint for the `z`-dimension).

    '''
    # Read in parameters:
    y_m = phase_map.phase.reshape(-1)  # Measured phase map as a vector
    a = phase_map.a  # Grid spacing
    dim = mask.shape  # Dimensions of the mag. distr.
    count = mask.sum()  # Number of pixels with magnetization
    lam = 1e-6  # Regularisation parameter
    # Create empty MagData object for the reconstruction:
    mag_data_rec = MagData(a, (np.zeros(dim), np.zeros(dim), np.zeros(dim)))

    # Function that returns the phase map for a magnetic configuration x:
    def F(x):
        mag_data_rec.set_vector(mask, x)
        phase = pm.phase_mag_real(a, pj.simple_axis_projection(mag_data_rec), b_0)
        return phase.reshape(-1)

    # Cost function which should be minimized:
    def J(x_i):
        y_i = F(x_i)
        term1 = (y_i - y_m)
        term2 = lam * x_i
        return np.concatenate([term1, term2])

    # Reconstruct the magnetization components:
    x_rec, _ = leastsq(J, np.zeros(3*count))
    mag_data_rec.set_vector(mask, x_rec)
    return mag_data_rec
